/*
 * int070atft-ts.c
 *
 *  Created on: 11.10.2012
 *      Author: emlin
 */

#define DEBUG

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/printk.h>

#include <asm/uaccess.h>

#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/delay.h>

#include <mach/hardware.h>
#include <mach/at91_pio.h>
#include <mach/gpio.h>

#include "int070atft-ts.h"
#include "ssd1963_cmd.h"

// -----------------------------------------------------------------------------
#define DEV_NAME				"int070atft-ts"
#define DEV_COUNT				1

#define AT91_PIOC_BASE			(AT91_PIOC + AT91_BASE_SYS)
#define	AT91_PIOC_SIZE			512

// -----------------------------------------------------------------------------
enum
{
	ModuleExit = 0,

	Failed_ChrdevRegion,
	Failed_CdevAlloc,
	Failed_CdevAdd,
	Failed_SysClass,
	Failed_SysDev,
	Failed_ReqRegion,
	Failed_IoMap,
};

// -----------------------------------------------------------------------------
// parametry ovladace
struct int070atft_t
{
	// major-minor par
	dev_t				dev_num;
	// ukazatel na strukturu cdev reprezentujici znakove zarizeni
	struct cdev *		cdev_ptr;
	// ukazatel na tridu zarizeni v adresari /sys/class
	struct class *		sys_class_ptr;
	struct device * 	sys_dev_ptr;
	// hlida pocet volani operace open()
	atomic_t			open_cnt;
	// zacatek registru PIOB brany jako I/O pamet
	void __iomem *		pio_base;
	// buffer znaku
};

static struct int070atft_t tft =
{
	.dev_num = MKDEV(0, 0),
	.cdev_ptr = NULL,
	.sys_class_ptr = NULL,
	.sys_dev_ptr = NULL,
	.open_cnt = ATOMIC_INIT(0),
	.pio_base = NULL,
};

// -----------------------------------------------------------------------------
// Pozn.: vychazim z toho, ze jedna instrukce pri 200MHz taktu trva 5ns
static void write_delay(void)
{
	// cca 15ns
	asm volatile
	(
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
	);
}

static inline void read_delay(void)
{
	// cca 35ns
	asm volatile
	(
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
		"NOP\n\t"
	);
}

// -----------------------------------------------------------------------------
/*
 * Pozn.: Pozor neobsahuji pametove bariery (rmb, wmb, mb())!
 */
static inline unsigned int pio_readl(unsigned int reg_offset)
{
	return __raw_readl(tft.pio_base + reg_offset);
}

static inline void pio_writel(unsigned int reg_offset, unsigned int value)
{
	__raw_writel(value, (tft.pio_base + reg_offset));
}

// -----------------------------------------------------------------------------
/*
 * prepina vodice 16bit sbernice pro TFT na cteni (PIO na input) nebo zapis
 * (PIO na vystup)
 */
static inline void set_bus_read(void)
{
	pio_writel(PIO_ODR, TFT_DB0_DB15_PIN_MASK);
}

static inline void set_bus_write(void)
{
	pio_writel(PIO_OER, TFT_DB0_DB15_PIN_MASK);
}

// -----------------------------------------------------------------------------
static inline unsigned int read_raw(void)
{
	// orizneme na 16bit a posuneme zapisovanou hodnotu na PC16 - PC31
	unsigned int reg_val;

	// shodime RD
	at91_set_gpio_value(TFT_nRD, 0);

	// pockame min 36ns
	read_delay();

	// SSD1963 vystavi na D0-D15 data, vycteme si je
	reg_val = pio_readl(PIO_PDSR);

	// ukoncime cteni dat vytazenim RD na log. 1
	at91_set_gpio_value(TFT_nRD, 1);

	return (reg_val >> 16);
}

// -----------------------------------------------------------------------------
static inline void write_raw(unsigned int value)
{
	// orizneme na 16bit a posuneme zapisovanou hodnotu na PC16 - PC31
	unsigned int reg_val = (value & 0xFFFF) << 16;

	// shodime WR
	at91_set_gpio_value(TFT_nWR, 0);

	// na D0-D15 vystavime data
	pio_writel(PIO_ODSR, reg_val);

	// pockame min 12ns
	write_delay();

	// zapiseme data vytazenim WR na log. 1
	at91_set_gpio_value(TFT_nWR, 1);
}

// -----------------------------------------------------------------------------
static inline unsigned int read_data(void)
{
	unsigned int data;

	// nastavime signal RS na data
	at91_set_gpio_value(TFT_DC, 1);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// vycteme data
	data = read_raw();

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);

	return data;
}

// -----------------------------------------------------------------------------
static inline void write_data(unsigned int data)
{
	// nastavime signal D/C (RS) na data
	at91_set_gpio_value(TFT_DC, 1);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// zapiseme data
	write_raw(data);

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);
}
// -----------------------------------------------------------------------------
static inline void write_data_rep(unsigned int data, const unsigned int repcnt)
{
	int i;

	// nastavime signal RS na data
	at91_set_gpio_value(TFT_DC, 1);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	for ( i = 0; i < repcnt; i++ )
	{
		// zapiseme data
		write_raw(data);
	}

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);
}

// -----------------------------------------------------------------------------
static inline void write_cmd(unsigned char cmd)
{
	// nastavime signal RS na data
	at91_set_gpio_value(TFT_DC, 0);

	// aktivujeme chipselect
	at91_set_gpio_value(TFT_nCS, 0);

	// zapiseme cmd
	write_raw(cmd);

	// deaktivujeme chipset
	at91_set_gpio_value(TFT_nCS, 1);
}

// -----------------------------------------------------------------------------
static void soft_reset(void)
{
	set_bus_write();
	write_cmd(CMD_SOFT_RESET);

	// must wait for 5ms at least
	msleep(5);
}

// -----------------------------------------------------------------------------
static void set_pll_mn(void)
{
	set_bus_write();
	write_cmd(CMD_SET_PLL_MN);
	// M = 49, N = 3, Xtal = 8 MHz
	// PLL = Xtal * (M + 1) / (N + 1) = 100 MHz
	write_data(0x31);
	write_data(0x03);
	// potvrdime nastaveni M a N
	write_data(0x04);
}

// -----------------------------------------------------------------------------
static void start_pll(void)
{
	set_bus_write();
	// zapneme PLL
	write_cmd(CMD_PLL_START);
	write_data(0x01);

	// pockame 100us na stabilizaci PLL
	udelay(100);

	// nastavime PLL jako systemove hodiny
	write_cmd(CMD_PLL_START);
	write_data(0x03);
}

// -----------------------------------------------------------------------------
static void set_lshift_freq(void)
{
	// nastavuji pixel clock (rychlost vykreslovani pixelu za sec)
	// defakto pak urcuje spolu s parametry hsync a vsync refresh rate displaye
	set_bus_write();
	write_cmd(CMD_SET_PCLK);
	// LCDC_FPR = 0x4cccc, PLL = 100MHz
	// PCLK = PLL * (LCDC_FPR + 1) / 2^20 = 30 Mhz
	write_data(0x04);
	write_data(0xcc);
	write_data(0xcc);
}

// -----------------------------------------------------------------------------
static void set_lcd_mode(void)
{
	set_bus_write();
	write_cmd(CMD_SET_LCD_MODE);
	// [5]=0: 18bit rozhrani k LCD panelu,
	// [4]=1: enable FRC or dithering, [3]=1: FRC enable,
	// [2]=0: lshift latch on falling edge, [1]=0: hsync active low, [0]=0: vsync active low
	write_data(0x18);
	// TFT mode
	write_data(0x00);
	// 800 pixelu na sirku HDP = 800 - 1
	write_data(0x03);
	write_data(0x1f);
	// 480 pixelu na vysku VDP = 480 - 1
	write_data(0x01);
	write_data(0xdf);
	// RGB pro oboje
	write_data(0x00);
}

// -----------------------------------------------------------------------------
static void set_address_mode(void)
{
	set_bus_write();
	write_cmd(CMD_SET_ADDR_MODE);
	// [7]=0: pages top->bottom, [6]=0: columns left->right, [5]=0: columns normal mode
	// [4]=0: refresh top->bottom, [3]=0: RGB->RGB, [2]=0: refresh left->right,
	// [1]=0: flip horizontal->normal, [0]=0: flip vertical->normal
	write_data(0x00);
}

// -----------------------------------------------------------------------------
static void set_pixel_data_interface(void)
{
	set_bus_write();
	write_cmd(CMD_SET_DATA_INTERFACE);
	// [2:0]=011: sbernice 16bitova, data prenasena v jednom cyklu(RGB565)
	write_data(0x03);
}

// -----------------------------------------------------------------------------
static void set_pixel_format(void)
{
	/* function: defines how many bits per pixel is used
	 * description:
	 * v ofiko docu je tento prikaz jako reserved, nicmene v dostupnych
	 * zdrojacich je pouzivan dle nasledujici notace:
	 * Display pixel format[6:4] (POR = 000):
	 * 			000 Reserved
	 * 			001 3-bit/pixel
	 * 			010 8-bit/pixel
	 * 			011 12-bit/pixel
	 * 			100 Reserved
	 * 			101 16-bit/pixel
	 * 			110 18-bit/pixel
	 * 			111 24-bit/pixel
	 */
	set_bus_write();
	write_cmd(CMD_SET_PIXEL_FORMAT);
	// [6:4]=101: 16bit/pixel protoze pouzivame 16bit, RGB565 rozhrani
	write_data(0x05 << 4);
}

// -----------------------------------------------------------------------------
static void set_hori_period(void)
{
	set_bus_write();
	write_cmd(CMD_SET_HOR_PERIOD);
	// HT = 0x0434 (1075)
	// celkova doba pro vykresleni radku (hsync pulse, front a back porch a samotne pixely)
	write_data(0x04);
	write_data(0x34);
	// HPS = 0x0002 (2)
	//
	write_data(0x00);
	write_data(0x02);
	// HPW = 0x15 (21)
	write_data(0x15);
	// LPS = 0x000000
	write_data(0x00);
	write_data(0x00);
	write_data(0x00);
}

// -----------------------------------------------------------------------------
static void set_vert_period(void)
{
	set_bus_write();
	write_cmd(CMD_SET_VER_PERIOD);
	// VT = 0x0221 (545)
	write_data(0x02);
	write_data(0x21);
	// VPS = 0x0016 (22)
	write_data(0x00);
	write_data(0x16);
	// VPW = 0x15 (21)
	write_data(0x15);
	// FPS = 0x000000
	write_data(0x00);
	write_data(0x00);
}

// -----------------------------------------------------------------------------
static void set_column_address(unsigned int sc, unsigned int ec)
{
	set_bus_write();
	write_cmd(CMD_SET_COLUMN);
	// start column address (SC)
	write_data(sc >> 8);
	write_data(sc);
	// end column address (EC)
	write_data(ec >> 8);
	write_data(ec);
}

// -----------------------------------------------------------------------------
static void set_page_address(unsigned int sp, unsigned int ep)
{
	set_bus_write();
	write_cmd(CMD_SET_PAGE);
	// start page address (SP)
	write_data(sp >> 8);
	write_data(sp);
	// end page address (EP)
	write_data(ep >> 8);
	write_data(ep);
}

// -----------------------------------------------------------------------------
static void set_gpio_value(void)
{
	set_bus_write();
	write_cmd(CMD_SET_GPIO_VAL);
	// nastavime log. 1
	write_data(0x0f);
}

// -----------------------------------------------------------------------------
static void set_gpio_conf(void)
{
	set_bus_write();
	write_cmd(CMD_SET_GPIO_CONF);
	// [7:4]=0: GPIO rizeno MCU, [3:0]=1: GPIO jsou vystupy
	write_data(0x0f);
	// [0]=1: GPIO0 je pouzito jako normalni GPIO ne pro rizeni napajeni
	write_data(0x01);
}

// -----------------------------------------------------------------------------
static void set_display_on(void)
{
	set_bus_write();
	write_cmd(CMD_ON_DISPLAY);
}

// -----------------------------------------------------------------------------
static void set_pwm_conf(unsigned char brightness)
{
	set_bus_write();
	write_cmd(CMD_SET_PWM_CONF);

	// PWM frekvence (PWMclk ~ 300Hz)
	// PWMF = PLL / (256^2 * PWMclk) = 100e6 / (256^2 * 300) = 5,0863 ~ 5
	write_data(0x05);
	// modulacni pomer (0x00 (tma) - 0xFF(max. jas))
	write_data(brightness);
	// zapneme PWM a nastavime rizeni z MCU (ne pomoci DBC)
	write_data(0x01);
	// max. jas, kdyz to ridi DBC
	write_data(0x00);
	// min. jas, kdyz to ridi DB
	write_data(0x00);
	// krok zmeny jasu - vypnuto
	write_data(0x00);
}

// -----------------------------------------------------------------------------
static void set_dbc_conf(void)
{
	set_bus_write();
	write_cmd(CMD_SET_ABC_DBC_CONF);

	// aktivuji a konfiguruji dynamic backlight control (DBC)
	// [6]=0: DBC manual brightness enable
	// [5]=0: Transition effect disable
	// [3:2]=11: aggressive mode
	// [0]=1: DBC enable
	write_data(0x0d);

	// Pozn.: To ale jeste neznamena, ze PWM bude rizeno DBC!
	// PWM se konfiguruje pred DBC! Viz popisek k prikazu set_dbc_conf
}

// -----------------------------------------------------------------------------
static void set_post_proc(void)
{
	set_bus_write();
	write_cmd(CMD_SET_POST_PROC);

	// Contrast 0 - 255
	write_data(120);
	// Brightness 0x00 - 0xFF
	write_data(100);
	// Saturation 0x00 - 0xFF
	write_data(100);
	// zapneme post processing obrazu
	write_data(0x01);
}

// -----------------------------------------------------------------------------
static void tft_clear(unsigned int color)
{
	set_column_address(TFT_SC, TFT_EC);
	set_page_address(TFT_SP, TFT_EP);

	// dame SSD1963 vedet, ze budeme ted zapisovat do framebufferu
	write_cmd(CMD_WR_MEMSTART);
	// zapiseme kod barvy na celou plochu displeje
	write_data_rep(color, (TFT_ROWS * TFT_COLS));
}

// -----------------------------------------------------------------------------
static void __init init_pio(void)
{
	// vypneme u PIO pro LCD preruseni
	pio_writel(PIO_IDR, (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// vypneme u PIO pro LCD pullup odpory
	pio_writel(PIO_PUDR, (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// nastavime ridici signaly na log. 1
	pio_writel(PIO_SODR, TFT_CTRL_PINS_MASK);
	// nastavime data na log. 0 (prikaz NOP)
	pio_writel(PIO_CODR, TFT_DB0_DB15_PIN_MASK);
	// prepneme PIO pro TFT na vystupni PIO
	pio_writel(PIO_OER, (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
	// pro D0-D17 povolime zapis dat primo pres PIO_ODSR
	// lze tak zapisovat rovnou cele slovo, nicmene prenesou se jen ty bity,
	// ktere maji v masce 1 v PIO_OWER
	pio_writel(PIO_OWER, TFT_DB0_DB15_PIN_MASK);
	// prepneme piny pro TFT na funkci GPIO
	pio_writel(PIO_PER, (TFT_CTRL_PINS_MASK | TFT_DB0_DB15_PIN_MASK));
}

// -----------------------------------------------------------------------------
static void __init init_tft(void)
{
	// mozna nahradit zatahanim za HW reset? Nebo ten bude pripojen k NRST MCU?
	soft_reset();
	// nastaveni a aktivace PLL jao systemovych hodin
	set_pll_mn();
	start_pll();
	// znovu SW reset
	soft_reset();
	// nastaveni pixel clock - rychlost vykreslovani obrazovych bodu
	set_lshift_freq();
	// nastavujeme parametry TFT panelu a zpusob adresace framebuferu
	set_lcd_mode();
	set_address_mode();
	// nastavujeme sirku datove sbernice a barevnou hloubku/schema
	set_pixel_data_interface();
	set_pixel_format();
	// nasatvujeme hsync a vsync parametry
	set_hori_period();
	set_vert_period();
	// nastavujeme zobrazovanou plochu
	set_column_address(TFT_SC, TFT_EC);
	set_page_address(TFT_SP, TFT_EP);
	// GPIO piny jako vystupy a do log. 1
	set_gpio_value();
	set_gpio_conf();
	// prepneme display do stavu DISPLAY
	set_display_on();

	// 0 (tma) - 255 (max jas)
	set_pwm_conf(180);
	// PWM tim rizeno neni
	set_dbc_conf();
	// nastavime kontarst, jas a stauraci
	set_post_proc();
	// bila na cely displej
	tft_clear(0xffff);
}

// -----------------------------------------------------------------------------
static int int070atft_open(struct inode * inode_ptr, struct file * file_ptr)
{
	int result = 0, tmp_open_cnt = 0;

	// soubor zarizeni muze byt otevren jen jednim programem soucasne
	// atomicky inkrementuje promenou o 1 a vrati PUVODNI hodnotu
	tmp_open_cnt = atomic_inc_and_test(&tft.open_cnt);
	if ( tmp_open_cnt > 1 )
	{
		pr_err(DEV_NAME"_open(): already opened!\n");
		atomic_dec(&tft.open_cnt);
		result = -EBUSY;
		goto End;
	}

	// modul nepodporuje systemove volani lseek()
	result = nonseekable_open(inode_ptr, file_ptr);
	if ( result < 0 )
	{
		goto End;
	}

End:
	pr_devel(DEV_NAME"_open(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int int070atft_release(struct inode * inode_ptr, struct file * file_ptr)
{
	// povolime nove otevreni
	atomic_dec(&tft.open_cnt);

	pr_devel(DEV_NAME"_release(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
static ssize_t int070atft_read(struct file * file_ptr,
				   				char __user * data_ptr, size_t data_size,
				   				loff_t * pos_ptr)
{
	pr_devel(DEV_NAME"_read(): failed\n");

	return -EINVAL; // nelze cist;
}

// -----------------------------------------------------------------------------
static ssize_t int070atft_write(struct file * file_ptr,
				    			const char __user * data_ptr, size_t data_size,
				    			loff_t * pos_ptr)
{
	ssize_t result = data_size;

	if ( data_size > 0 )
	{
		unsigned int color = 0x0;
		long res;

		res = copy_from_user(&color, data_ptr, TFT_COLOR_DEPTH_BYTES);
		if ( res < 0 )
		{
			result = -EFAULT;
			goto End;

		}

		tft_clear(color);
	}

End:
	pr_devel(DEV_NAME"_write(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
// virtualni tabulka - tabulka operaci podporovanych ovladacem
static struct file_operations int070atft_file_ops =
{
	.owner = THIS_MODULE,
	.open = &int070atft_open,
	.release = &int070atft_release,
	.read = &int070atft_read,
	.write = &int070atft_write,
	// modul nepodporuje metodu llseek
	.llseek = &no_llseek,
};

// -----------------------------------------------------------------------------
static void cleanup(int error)
{
	switch ( error )
	{
		// kompletni uklid pri exitu
		case ModuleExit:
			// zrusime adresar modulu v /sys
			device_destroy(tft.sys_class_ptr, tft.dev_num);

		case Failed_SysDev:
			// zrusime tridu
			class_destroy(tft.sys_class_ptr);

		case Failed_SysClass:
			// vyjmeme modul z jadra
			// pozn.: cdev_del() se vola jen po uspesnem cdev_add()
			cdev_del(tft.cdev_ptr);

		case Failed_CdevAdd:
			// zrusime mapovani I/O pameti do adresoveho prostoru jadra
			iounmap(tft.pio_base);

		case Failed_IoMap:
			// zrusime rezervaci I/O portu
			release_mem_region(AT91_PIOC_BASE, AT91_PIOC_SIZE);

		case Failed_ReqRegion:
			// pozn.: neexistuje opak fce cdev_alloc(), uvolneni pameti resi
			// jadro samo
		case Failed_CdevAlloc:
			// uvolnime nami zabrany major-minor par
			unregister_chrdev_region(tft.dev_num, DEV_COUNT);

		case Failed_ChrdevRegion:
			// jeste nic se nealokovalo, takze neni co vracet
			break;

		default:
			break;
	}
}

// -----------------------------------------------------------------------------
static int __init int070atft_init(void)
{
	int result = 0;
	struct resource * region_ptr;

	// registrace k major-minor paru
	result = alloc_chrdev_region(&tft.dev_num, MINOR(tft.dev_num),
								 DEV_COUNT, DEV_NAME);
	if ( result < 0 )
	{
		cleanup(Failed_ChrdevRegion);
		goto End;
	}

	// alokace pameti pro strukturu cdev
	tft.cdev_ptr = cdev_alloc();
	if ( tft.cdev_ptr == NULL )
	{
		result = -ENOMEM;
		cleanup(Failed_CdevAlloc);
		goto End;
	}

	// nastaveni vlastnika a seznamu souborovych operaci
	tft.cdev_ptr->owner = THIS_MODULE;
	tft.cdev_ptr->ops = &int070atft_file_ops;

	// rezervujeme si PIO port
	region_ptr = request_mem_region(AT91_PIOC_BASE, AT91_PIOC_SIZE, DEV_NAME);
	if ( region_ptr == NULL )
	{
		pr_err(DEV_NAME"_init(): cannot reserve PIOC!\n");
		cleanup(Failed_ReqRegion);
		result = -EBUSY;
		goto End;
	}

	// premapujeme PIO port do adresoveho prostoru jadra
	tft.pio_base = ioremap(AT91_PIOC_BASE, AT91_PIOC_SIZE);
	if ( tft.pio_base == NULL )
	{
		cleanup(Failed_IoMap);
		result = -ENOMEM;
		goto End;
	}

	// inicializujeme LCD displej pro prvni pouziti
	// nastaveni rezimu prenosu, vstupniho rezimu apod.
	init_pio();
	init_tft();

	//  --- pridani ovladace do jadra ---
	result = cdev_add(tft.cdev_ptr, tft.dev_num, DEV_COUNT);
	if ( result < 0 )
	{
		cleanup(Failed_CdevAdd);
		goto End;
	}

	// vytvorime si tridu v /sys/class
	tft.sys_class_ptr = class_create(THIS_MODULE, DEV_NAME);
	if ( IS_ERR(tft.sys_class_ptr) )
	{
		cleanup(Failed_SysClass);
		result = PTR_ERR(tft.sys_class_ptr);
		goto End;
	}

	// pridame nas modul do /sys/class/int070atft-ts
	tft.sys_dev_ptr = device_create(tft.sys_class_ptr, NULL, tft.dev_num,
									NULL, DEV_NAME);
	if ( IS_ERR(tft.sys_dev_ptr) )
	{
		cleanup(Failed_SysDev);
		result = PTR_ERR(tft.sys_dev_ptr);
		goto End;
	}

End:
	pr_devel(DEV_NAME"_init(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static void __exit int070atft_exit(void)
{
	cleanup(ModuleExit);

	pr_devel(DEV_NAME"_exit(): done\n");
}

// -----------------------------------------------------------------------------
module_init(int070atft_init);
module_exit(int070atft_exit);

MODULE_AUTHOR("mBoss <marek.laznicka@elvoris.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for Displaytech INT070ATFT-TS TFT-LCD display");
MODULE_VERSION("0.1.0");
